
import java.io.File;
import java.io.FileFilter;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.concurrent.*;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;

/**
 * Threaded Find Duplicates.
 * Modification of ProducerConsumer
 * Chapter 5.3 of "Java Concurrency in Practice"
 * 
 * @author Nicolás Wolovick
 */

public class ThreadedFindDuplicates {
    static class FileCrawler implements Runnable {
    	private final BoundedQueue<File> fileQueue;
        private final FileFilter fileFilter;
        private final File root;

        //http://www.idiom.com/~zilla/Xfiles/javasymlinks.html
        public static boolean isLink(File file)
        {
        	try {
        		if (!file.exists())
        			return true;
        		else {
        			String cnnpath = file.getCanonicalPath();
        			String abspath = file.getAbsolutePath();
        			return !abspath.equals(cnnpath);
        		}
        	}
        	catch(IOException ex) {
        		System.err.println(ex);
        		return true;
        	}
        }
        
        public FileCrawler(BoundedQueue<File> fileQueue,
                           final FileFilter fileFilter,
                           File root) {
            this.fileQueue = fileQueue;
            this.root = root;
            this.fileFilter = new FileFilter() {
                public boolean accept(File f) {
                    return !isLink(f) && 0<f.length() && f.canRead() && (f.isDirectory() || fileFilter.accept(f));
                }
            };
        }

        private boolean alreadyIndexed(File f) {
        	// does not handle cycles in the directory structure
            return false;
        }

        public void run() {
            try {
                crawl(root);
            } catch (InterruptedException e) {
                Thread.currentThread().interrupt();
            }
        }

        private void crawl(File root) throws InterruptedException {
            File[] entries = root.listFiles(fileFilter);
            if (entries != null) {
                for (File entry : entries)
                    if (entry.isDirectory())
                        crawl(entry);
                    else if (!alreadyIndexed(entry))
                        fileQueue.put(entry);
            }
        }
    }

    static class Checker implements Runnable {
        private final BoundedQueue<File> queue;
        private final ConcurrentHashMap<String,String> fileMD5Map;

        public Checker(BoundedQueue<File> queue, ConcurrentHashMap<String,String> fileMD5Map) {
            this.queue = queue;
			this.fileMD5Map = fileMD5Map;
        }

        public void run() {
        	String md5 = null;
        	String oldFilename;
        	File file;
            try {
            	while (true) {
            		file = queue.take();
            		assert file.isFile() && file.canRead();
            		try {
            			md5 = this.md5file(file);
            		} catch (Exception e) {
            			System.err.println("Couldn't compute MD5 for "+file.toString()+": "+e);
            		}
            		oldFilename = fileMD5Map.putIfAbsent(md5, file.toString());
            		if (oldFilename!=null && !oldFilename.equals(file.toString())) {
            			System.out.println(file.toString()+" is duplicate of "+oldFilename); 
            		}
            	}
            } catch (InterruptedException e) {
                Thread.currentThread().interrupt();
            }
        }

        private String md5file(File file) throws NoSuchAlgorithmException, IOException {
        	//http://www.oreilly.com/catalog/javacrypt/chapter/ch06.html
			MessageDigest algorithm = MessageDigest.getInstance("MD5");
        	InputStream is = new FileInputStream(file);
        	byte[] buffer = new byte[8192];
        	int read = 0;
        	algorithm.reset();
        	try {
        		while( (read = is.read(buffer)) > 0) {
        			algorithm.update(buffer, 0, read);
        		}
        		byte md[] = algorithm.digest();
        		StringBuffer hexString = new StringBuffer();
        		for (int i=0;i<md.length;i++) {
        			hexString.append(Integer.toHexString(0xFF & md[i]));
        		}

        		return hexString.toString();
        	}
        	finally {
        			is.close();
        	}		
        }
        
    };

    private static final int CAPACITY = 16;
    private static final int N_CONSUMERS = Runtime.getRuntime().availableProcessors();

    
    public static void main(String[] args) {
        BoundedQueue<File> queue = new BoundedQueue<File>(CAPACITY);
        ConcurrentHashMap<String,String> md5FileMap = new ConcurrentHashMap<String,String>();
        FileFilter filter = new FileFilter() {
            public boolean accept(File file) {
                return true;
            }
        };

        for (String arg : args) {
            new Thread(new FileCrawler(queue, filter, new File(arg)), "Crawler for "+arg).start();
        	System.out.println("Spawned crawler for: " + arg);
        }
        for (int i = 0; i < N_CONSUMERS; i++) {
            new Thread(new Checker(queue,md5FileMap), "Checker "+i).start();
        	System.out.println("Spawned checker " + i);
        }
    }
}